/*
 * Copyright (C) 2006-2010 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file stdout.c Functions that deal with writing to stdout. */

#include <dirent.h> /* readdir() */
#include <assert.h>
#include <stdbool.h>

#include "para.h"
#include "list.h"
#include "sched.h"
#include "fd.h"
#include "error.h"
#include "stdout.h"
#include "buffer_tree.h"

/**
 * The pre_select function of the stdout task.
 *
 * \param s The scheduler this task was registered to.
 * \param t The task structure of the stdout task.
 *
 * This function is always successful. If there is input data available, it
 * adds \p STDOUT_FILENO to the write fd set of \a s.
 */
static void stdout_pre_select(struct sched *s, struct task *t)
{
	struct stdout_task *sot = container_of(t, struct stdout_task, task);
	int ret;

	t->error = 0;
	ret = btr_node_status(sot->btrn, 0, BTR_NT_LEAF);
	if (ret > 0)
		para_fd_set(STDOUT_FILENO, &s->wfds, &s->max_fileno);
	else if (ret < 0)
		sched_min_delay(s);
}

/**
 * The post select function of the stdout task.
 *
 * \param s The scheduler this task was registered to.
 * \param t The task structure of the stdout task.
 *
 * This function writes input data from the buffer tree to stdout if \p
 * STDOUT_FILENO is writable.
 */
static void stdout_post_select(struct sched *s, struct task *t)
{
	struct stdout_task *sot = container_of(t, struct stdout_task, task);
	struct btr_node *btrn = sot->btrn;
	int ret;
	char *buf;
	size_t sz;

	t->error = 0;
	ret = btr_node_status(btrn, 0, BTR_NT_LEAF);
	if (ret < 0)
		goto out;
	if (ret == 0)
		return;
	if (!FD_ISSET(STDOUT_FILENO, &s->wfds))
		return;

	for (;;) {
		sz = btr_next_buffer(btrn, &buf);
		if (sz == 0)
			break;
		ret = write_nonblock(STDOUT_FILENO, buf, sz);
		if (ret <= 0)
			break;
		btr_consume(btrn, ret);
	}
out:
	if (ret < 0)
		btr_remove_node(btrn);
	t->error = ret;
}
/**
 * Initialize a stdout task structure with default values.
 *
 * \param sot The stdout task structure.
 *
 * This fills in the pre/post select function pointers of the task structure
 * given by \a sot and sets the stdout file descriptor to nonblocking mode.
 */
void stdout_set_defaults(struct stdout_task *sot)
{
	int ret;

	sot->task.pre_select = stdout_pre_select;
	sot->task.post_select = stdout_post_select;
	sprintf(sot->task.status, "stdout");
	ret = mark_fd_nonblocking(STDOUT_FILENO);
	if (ret >= 0)
		return;
	PARA_EMERG_LOG("%s\n", para_strerror(-ret));
	exit(EXIT_FAILURE);
}
